/*
 * @Author: WangJiaCheng
 * @Date: 2021-06-22 17:10:11
 * @LastEditors: LiZhiWei
 * @LastEditTime: 2025-04-16 09:53:14
 * @Description: 文件相关工具函数
 */
import { Notification, Message } from 'element-ui'
import { saveAs } from 'file-saver'

import axios from './api.request'

type FileInfo = {
  name: string
  size: string
}

/**
 * 文件下载
 * @author WangJiaCheng
 * @param {string} obj  下载地址 | Blob
 * @param {string} url  文件名，可选的
 */
export function saveDownload (obj: string, fileName?: string) {
  try {
    saveAs(obj, fileName)
  } catch (e) {
    console.log(e)
  }
}

/**
 * 进度条下载，使用Blob类型下载
 * @param {string} url 文件下载地址
 * @param {vnode} $vm 组件实例，使用中即是this
 * @param {name, size} fileInfo 文件信息，包括文件名、文件大小
 */
export function downloadWithProgress (url: string, _vm, loading: string, fileInfo: FileInfo = { name: '', size: '' }, callBack) {
  let p = 1
  let progressTimer
  const { name, size } = fileInfo
  const notificationInstance = Notification({
    title: '文件下载中，请稍候',
    offset: 50,
    customClass: 'notification-down-progress',
    message: _vm.$createElement('el-progress', {
      style: {
        width: '280px'
      },
      props: {
        percentage: p,
        textInside: true,
        strokeWidth: 18
      },
      ref: 'downloadProgressBar'
    }),
    duration: 0
  })
  const settingProgressBar = (p) => {
    const progressBar = document.querySelector('.notification-down-progress .el-progress-bar__inner') as HTMLElement
    const progressBarText = document.querySelector('.notification-down-progress .el-progress-bar__innerText') as HTMLElement
    if (progressBarText && p) {
      progressBarText.innerHTML = `${p}%`
      progressBar.style.width = `${p}%`
      progressBar.style.backgroundColor = '#409eff'
    }
  }
  // 在等待后端响应blob数据前，创建虚拟的下载百分比，给用户反馈下载进度
  if (notificationInstance) {
    progressTimer = setInterval(() => {
      settingProgressBar(p + 1)
    }, 800)
  }
  axios.request({
    url,
    method: 'get',
    params: { wc: new Date().getTime() }, // 后端默认需要字段
    responseType: 'blob',
    headers: {
      Accept: 'application/json, text/plain, */*'
    },
    withCredentials: false,
    timeout: 1000 * 60 * 10, // 10分钟
    onDownloadProgress: (event) => {
      const total = event.total || size
      if (event) {
        clearInterval(progressTimer)
      }
      if (event.target.status >= 200 && event.target.status < 300) {
        p = Math.round((event.loaded * 100) / total)
        if (total) {
          settingProgressBar(p)
        }
        if (p >= 100) {
          setTimeout(() => {
            notificationInstance.close()
          }, 500)
        }
      } else {
        notificationInstance.close()
      }
    }
  }).then((res) => {
    settingProgressBar(100)
    if (res.data) {
      const downFileName = name || res.fileName
      if (downFileName) {
        saveDownload(res.data, downFileName)
        _vm[loading] = false
      } else {
        saveDownload(url)
      }
      callBack()
    }
  }).catch((e) => {
    notificationInstance.close()
    _vm[loading] = false
    const reader = new FileReader()
    reader.onload = () => {
      const { result } = reader
      const errorInfos = JSON.parse(result as string)
      const message = errorInfos.responseJSON.message || '下载异常，请稍后尝试'
      Message({
        type: 'error',
        message: message
      })
    }
    // reader.readAsText(e)
  })
}

/**
 * iframe 下载
 * @param {string} url 下载链接
 */
export function downloadFile (url: string) {
  const iframe = document.createElement('iframe')
  iframe.style.display = 'none'
  function iframeLoad () {
    const win = iframe.contentWindow as Window
    const doc = win.document
    if (win.location.href === url) {
      if (doc.body.childNodes.length > 0) {
        // response is error
      }
      if (iframe.parentNode) {
        iframe.parentNode.removeChild(iframe)
      }
    }
  }
  if ('onload' in iframe) {
    iframe.onload = iframeLoad
  }
  iframe.src = ''
  document.body.appendChild(iframe)

  if (iframe.contentWindow) {
    iframe.contentWindow.location.href = url
  }
}
